# Nonlinear-fit.py       I. Marti-Vidal.      IMPRS Blackboard Lectures 2011.
import numpy as np
import pylab as pl
# SCRIPT TO PERFORM A NONLINEAR FIT OF THE MODEL A*SIN(2pi*t/T) (PARAMS: A AND T)

# NOTE: THERE ARE MUCH MORE COMPLETE AND ROBUST LIBRARIES TO PERFORM NONLINEAR FITS
# (E.G. SCIPY.OPTIMIZE.LEASTSQ), BUT THIS SMALL SCRIPT PRETENDS TO SHOW THE BASICS OF THE 
# ALGORITHMS BEHIND ALL THESE LIBRARIES.

# NOTE 2: THE APPROACH IN THIS SCRIPT REQUIRES RATHER GOOD PRIORIS FOR THE PARAMETERS.
# THIS REQUIREMENT CAN BE RELAXED IF AN INITIAL SEARCH OF THE MINIMUM OF THE 
# CHISQ IS PERFORMED (FOR INSTANCE, USING LEVENBERG-MARQUARDT, SIMPLEX,...). I LEAVE 
# THIS OPTION UP TO YOU.

#########################################################################
datafile = "precess.dat"  # Name of the file with data (format is: time  angle  error)
prioris = [5.0,600.]      # Prioris of amplitude (deg) and period (days) of precession.
niter = 20                # Number of iterations (0 -> Only plot model and estimate Cov.)
DelMin = 0.01             # Minimum relative change in parameters to stop iterations.
#########################################################################


# THE AMPLITUDE WILL BE PARAMETER 0; THE PERIOD WILL BE PARAMETER 1.

##################
# PROGRAM STARTS #
##################


tp = 2.*np.pi

# Function that returns the model:
def Model(t,params):
 return params[0]*np.sin(tp*t/params[1])

# Gradient of the model with respect to Amplitude:
def GradAmp(t,params):
 return np.sin(tp*t/params[1])

# Gradient of the model with respect to Period:
def GradPer(t,params):
 return -params[0]*tp*t/params[1]**2.*np.cos(tp*t/params[1])


# Function to compute the Covariance matrix of the fitting parameters:

def CovMat(data, weight, params):
 times = data[:,0]
 F11 = np.sum(weight*np.power(GradAmp(times,params),2))
 F12 = np.sum(weight*GradAmp(times,params)*GradPer(times,params))
 F22 = np.sum(weight*np.power(GradPer(times,params),2))

 MatF = np.matrix([[F11,F12],[F12,F22]]) # This the F matrix (i.e., the inverse of Cov):
 return MatF.I   # Returns the covariance matrix




# Function to compute the vector of residuals (multiplied by the gradient of the model):

def resids(data, weight, params):
 times = data[:,0]
 resids = (data[:,1]-Model(times,params))
 R1 = np.sum(weight*resids*GradAmp(times,params))
 R2 = np.sum(weight*resids*GradPer(times,params))

 ChiSq = np.sum(resids*resids*weight) # Returns also the ChiSq as a bonus.
 return ChiSq, np.array([[R1],[R2]])  # These are the ChiSq and the (column) vector of residuals.




# PROCEED WITH THE MODEL FITTING:

# Read data:

fdata = open(datafile).readlines()
data = np.array([map(float,line.split()) for line in fdata[3:]]) 


# Convert errors into weights (i.e., 1./error^2):
weight = np.power(data[:,2],-2.)

# Fitting loop:

params = prioris
MatC = CovMat(data,weight,params)
ChiSq, res = resids(data,weight,params)

for i in range(niter):

 deltapar = MatC*res 
 
 if np.max(np.abs(deltapar.T/params))<DelMin: # If convergence achieved, quit loop.
  break
 else:   # If not, update the parameter estimates.
  params = np.array(deltapar.T + params)[0]
  print "\n Iteration ",i+1,". New parameters:"
  print "Amplitude: ",params[0],"+/-",np.sqrt(MatC[0,0])," deg."
  print "Period: ",params[1],"+/-",np.sqrt(MatC[1,1])," days."

 if i< niter:
  MatC = CovMat(data,weight,params)
  ChiSq, res = resids(data,weight,params)


# Once the loop is finished, print the best result:

sigA = np.sqrt(MatC[0,0]); sigT = np.sqrt(MatC[1,1])

print "\n\n FINAL VALUES OF THE PARAMETERS:\n"
print "Amplitude: ",params[0],"+/-",sigA," deg."
print "Period: ",params[1],"+/-",sigT," days.\n"
print " A-T correlation: ", MatC[0,1]/sigA/sigT," days*deg.\n"
print " Reduced ChiSq: ", ChiSq/(len(data)-2.),"\n"

# Finally, plot the fitted model and the data:

times = np.arange(data[0,0],data[-1,0])
pl.plot(data[:,0],data[:,1],'.')
pl.errorbar(data[:,0],data[:,1],data[:,2],fmt='k',linestyle='None')
pl.plot(times,params[0]*np.sin(tp*times/params[1]),'-r')
pl.xlabel("Time (days)");pl.ylabel("Angle (deg.)")
pl.title("LEAST-SQUARES FIT OF A PRECESSING JET")
pl.show()

# END
